/***
  *
  * Another implementation of Rough Sets which uses queries instead of
  * tables. Uses Shannon instead of original quality indicator.
  *
  * Copyright 2000,2001,2002 Robert H. Warren 
  * 
  * RSRules.java is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * RSRules.java is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  * on-line information : http://www.gnu.org/licenses/gpl.txt
  * 
  *
  * @Author Robert H. Warren
  * @Author rhwarren@uwaterloo.ca
  * @Version %Id:%
  * 
  * $Log: RSRules.java,v $
  * Revision 1.3  2002/10/11 03:46:14  warren
  * Fixed contact info.
  * Added license.
  *
  *
  * Revision 1.2  2000/09/14 19:39:06  warren
  * Production version using quality factor.
  *
  * Major problem with the quality factor is that the quality
  * factor fails when none of the columns contains a deterministic
  * (Upper - Lower) rule. Therefore, a backup column selection
  * attribute should be implemented.
  *
  * Revision 1.1  2000/09/13 20:38:59  warren
  * Initial revision
  *
 ***/
public class RSRules {
   /***
     * JDBC connection to the database in use.
    ***/
   java.sql.Connection TheDatabase = null;
   /***
     * Set to true if faulty trimming algorythm is to be used. Currently INOP.
    ***/
   boolean UniverseIsTrimmed = false;
   /***
     * Turns on global debugging.
    ***/
   boolean Debug = true;
   /***
     * Print SQL statements to the screen as they are sent to the DBMS.
    ***/
   boolean SDebug = true;
   /***
     * Stores the name of the column which represents the class of the row.
    ***/
   String ClassColumn = null;
   /***
     * Stores the name of the column which represents a unique row identifier.
    ***/
   String UNIQUEID = null;
   /***
     * Stores the name of the table being processed.
    ***/
   String InformationTable = null;
   /***
     * Stores all rules generated so far by the class.
    ***/
   java.util.Vector[] RuleChain = new java.util.Vector[2];
   /***
     * Internal constant to indicate that the rule is part of the lower bound.
    ***/
   final int LOWER =0;
   /***
     *Internal constant to indicate that the rule is part of the upper bound.
    ***/
   final int UPPER = 1;
   public static void main(String[] args) {
      try {
         Class.forName("org.postgresql.Driver");  
         RSRules myrs = new RSRules();
         myrs.TheDatabase = java.sql.DriverManager.getConnection("jdbc:postgresql://jill.tzone.org/research_alt","warren","mytools");
         myrs.setRowID("row_id");
         myrs.setClass("objective_function");
         myrs.setTable(args[0]);
         myrs.getRules();
      } catch (java.sql.SQLException e) {
         e.printStackTrace();
         System.err.println(e.toString());
      } catch (java.lang.ClassNotFoundException e) {
         e.printStackTrace();
         System.err.println(e.toString());
      } catch (java.lang.Exception e) {
         e.printStackTrace();
         System.err.println(e.toString());
      }    
   } //main 
   /***
     * Main processing loop. For each possible class value, get Rough Set rules and
     * dump them to the screen.
    ***/
   public void getRules() throws java.sql.SQLException {
      /* Get the number of classes */
      java.util.Vector ClassList = this.getValuesOfColumns(InformationTable, ClassColumn);
      if (Debug)
         System.err.println("RSRules.getRules: Will generate rules for class values: "+ClassList.toString());
      int AllRows = getRowCount(InformationTable);
      for (java.util.Enumeration DoClasses = ClassList.elements();DoClasses.hasMoreElements();) {
         String ClassValue = (String)DoClasses.nextElement();
         if (Debug)
            System.err.println("RSRules.getRules: Now generating rules for class ["+ClassValue+"].");
         java.util.Vector ColumnList = getTableColumns(InformationTable);
         ColumnList.removeElement(UNIQUEID);
         ColumnList.removeElement(ClassColumn);
         java.util.Vector[] Boundairy = generateRules(ColumnList, (new java.util.Vector()), InformationTable, ClassValue);
         java.util.Enumeration UpperBoundairy = Boundairy[UPPER].elements();
         java.util.Enumeration LowerBoundairy = Boundairy[LOWER].elements();
         System.out.println("Rules generated for class ["+ClassValue+"].");
         while (LowerBoundairy.hasMoreElements()) {
            RSBound UpperBound = (RSBound)UpperBoundairy.nextElement();
            RSBound LowerBound = (RSBound)LowerBoundairy.nextElement();
            if (UpperBound.Coverage!=0) {
               System.out.println("Rule ["+UpperBound.SQLFragment+"] coverage ["+(1.0*(UpperBound.Coverage+LowerBound.Coverage))/AllRows+"] support ["+1.0*LowerBound.Coverage/AllRows+"], boundairy width ["+((1.0*(UpperBound.Coverage+LowerBound.Coverage))/AllRows-1.0*LowerBound.Coverage/AllRows)+"].");                
            } else
               System.out.println("***Rule ["+UpperBound.SQLFragment+"] total coverage ["+(1.0*LowerBound.Coverage/AllRows)+"] with support ["+1.0*LowerBound.Coverage/AllRows+"].");                

         }
      }//for class
   }//getrules
   /***
     * Main rule generating recursive method. Will call itself each time a new column has been chosen
     * and iterates to find a new column to pick, storing rules as it goes.
     * @returns A vector of the rules generated so far.
     * @param ColumnsToDo A list of columns which have not been picked yet.
     * @param ColumnsDone A list of columns selected so far.
     * @param TableToUse Name of table being used within DBMS.
     * @param CurrentClass The class being used.
     *
    ***/
   public java.util.Vector[] generateRules(java.util.Vector ColumnsToDo, java.util.Vector ColumnsDone, String TableToUse, String CurrentClass) throws java.sql.SQLException {
      if (Debug)
         System.err.println("RSRules.generateRules: Need to pick pivot from ["+ColumnsToDo+"] with current pivot ["+ColumnsDone+"].");      
      double BestAlpha = 1.0;
      double BestShannon = 1.0;
      java.util.Vector[] BestShannonRuleSet = null;
      java.util.Vector[] BestRuleSet = null;      
      String BestColumn = null;
      String BestShannonColumn = null;
      java.util.Hashtable BoundData = new java.util.Hashtable();
      java.util.Vector AllColumns = getTableColumns(TableToUse);
      AllColumns.removeElement(UNIQUEID);
      AllColumns.removeElement(ClassColumn);
      String StatisticsTable = generateIndicernabilityFunction(AllColumns,TableToUse);
      java.sql.Statement CheckTable = TheDatabase.createStatement();
      String SQLStatement = new String("SELECT SUM(_CARDINAL) FROM "+StatisticsTable);
      if (SDebug)
         System.err.println("RSRules.generateRules: SQL ["+SQLStatement+"].");
      java.sql.ResultSet Counter = CheckTable.executeQuery(SQLStatement);
      Counter.next();
      int RowCount = Counter.getInt(1);
      if (Debug)
         System.err.println("RSRules.generateRules: There are currently ["+RowCount+"] rows in the system.");
      if (RowCount<= 0) {
         if (Debug)
            System.err.println("RSRules.generateRules: No more rows,exit recursing.");         
         return(null);
      }
      /* We try to use each column at a time. */
      for (java.util.Enumeration Targets = ColumnsToDo.elements();Targets.hasMoreElements();) {
         String ColumnTryout = (String)Targets.nextElement();
         if (Debug)
            System.err.println("RSRules.generateRules: Try column ["+ColumnTryout+"].");        
         java.util.Vector ColumnTryoutSet = (java.util.Vector)ColumnsDone.clone();
         ColumnTryoutSet.addElement(ColumnTryout);
         /* Two sets of bounds are returned for this order of columns. */
         java.util.Vector[] ThisBound = generateBounds(ColumnTryoutSet, StatisticsTable, CurrentClass);
         /* Calculate alpha factor... */
         int Weight = 0;
         double ShannonsWeight = getShannon(ColumnTryout, ColumnsDone, StatisticsTable, CurrentClass);
         java.util.Enumeration UpperCount = ThisBound[UPPER].elements();
         java.util.Enumeration LowerCount = ThisBound[LOWER].elements();
         if (Debug) {
            if (Weight==0)
               System.err.println("RSRules.generateRules: No deterministic rules, alpha measure fails.");
            else
               System.err.println("RSRules.generateRules: Some deterministic rules present.");
         }
         double MyAlpha = 1.0 - ((double) Weight/(double)RowCount);
         if (Debug)
            System.err.println("RSRules.generateRules: Alpha for this one is ["+MyAlpha+"].");   
         if (MyAlpha <= BestAlpha) {
            if (Debug)
               System.err.println("RSRules.generateRules: Best alpha is ["+BestAlpha+"].");
            BestRuleSet = ThisBound;
            BestColumn = ColumnTryout;
            BestAlpha = MyAlpha;
         }
         if (ShannonsWeight <= BestShannon) {
            BestShannon = ShannonsWeight;
            if (Debug)
               System.err.println("RSRules.generateRules: Best shannon is ["+BestShannon+"].");
            BestShannonRuleSet = ThisBound;
            BestShannonColumn = ColumnTryout;
         } //if
      } //for
      if (BestAlpha == 1.0) {
         if (Debug)
            System.err.println("RSRules.generateRules: The quality factor has Failed! Switching for shannon for this column selection.");
         BestColumn = BestShannonColumn;
         BestRuleSet = BestShannonRuleSet;
         if (Debug)
            System.err.println("RSRules.generateRules:  The new pivot column is ["+BestColumn+"].");
      } //if
      ColumnsToDo.removeElement(BestColumn);
      ColumnsDone.addElement(BestColumn);
      if (!ColumnsToDo.isEmpty()) {
         /* Clean up universe */
         String NewUniverse;
         if (UniverseIsTrimmed)
            NewUniverse = trimUniverse(TableToUse, BestRuleSet[UPPER]);
         else
            NewUniverse = TableToUse;

         /* Recurse call */
         java.util.Vector[] ChainedRuleSet = generateRules(ColumnsToDo, ColumnsDone, NewUniverse, CurrentClass);
         /* Merge rules */
         if (ChainedRuleSet!=null) {
            for (java.util.Enumeration TheList = ChainedRuleSet[0].elements();TheList.hasMoreElements();)
               BestRuleSet[0].addElement(TheList.nextElement());
            for (java.util.Enumeration TheList = ChainedRuleSet[1].elements();TheList.hasMoreElements();)
               BestRuleSet[1].addElement(TheList.nextElement());
         } //if
         /* Go backwards */
      } // No More Coluemns
      else {
         if (Debug)
            System.err.println("RSRules.generateRules: no more columns, exit recursion.");
      }
      return(BestRuleSet);
   }//generateRules
   /***
     * Broken universe trimming method. Currently INOP.
     *
     * @returns The name of the table within the DBMS containing the trimmed universe.
     * @param _input The name of the current universe table.
     * @param _rules The name of the rule chain to trim the universe with.
     *
    ***/
   public String trimUniverse(String _input, java.util.Vector _rules) throws java.sql.SQLException  {
      java.sql.Statement CheckTable = TheDatabase.createStatement();
      String _output = null;
      for (java.util.Enumeration Clauses = _rules.elements();Clauses.hasMoreElements();) {
         RSBound OneBoundairy =   (RSBound)Clauses.nextElement();
         if (OneBoundairy != null) {
            _output = generateTableName();
            String SQLStatement = new String("SELECT * INTO TEMP "+_output+" FROM "+_input+" WHERE NOT ("+OneBoundairy.SQLFragment+")");
            if (SDebug)
               System.err.println("RSRules.generateBounds: Sending SQL ["+SQLStatement+"].");
            if (SDebug)
               System.err.println("RSRules.generateBounds: Update returned ["+CheckTable.executeUpdate(SQLStatement)+"].");
            _input = _output;         
         } else
            _output = _input;
      }
      return(_output);
   } //trimUniverse
   /***
     * For all stated columns generate upper and lower bounds. 
     * @return A vector of the upper and lower bounds generated using the parameter columns.
     * @param ColumnValues The name of the column which will form the boundairy set.
     * @param _table The table to use as an universe.
     * @param _class The class value that is the target.
    ***/
   public java.util.Vector[] generateBounds(java.util.Vector ColumnValues, String _table, String _class) throws java.sql.SQLException {
      if (Debug)
         System.err.println("RSRules.generateBounds: Searching for class ["+_class+"] in table ["+_table+"] for columns ["+ColumnValues.toString()+"].");
      java.util.Vector[] Bounds = new java.util.Vector[2];
      Bounds[0] = new java.util.Vector();
      Bounds[1] = new java.util.Vector();
      /* Generate a column list for the query*/
      String Targets = new String();
      for (java.util.Enumeration AllTargets = ColumnValues.elements();AllTargets.hasMoreElements();)
         if (Targets.length() < 2)
            Targets = (String)AllTargets.nextElement();
         else
            Targets = new String(Targets+ ", "+(String)AllTargets.nextElement());
      /* For each variation in the table for the columns. */
      java.sql.Statement CheckTable = TheDatabase.createStatement();
      String SQLStatement = new String("SELECT DISTINCT " +Targets+ " FROM "+_table);
      if (SDebug)
         System.err.println("RSRules.generateBounds: Sending SQL ["+SQLStatement+"].");
      java.sql.ResultSet Combinations = CheckTable.executeQuery(SQLStatement);
      while (Combinations.next()) {
         RSBound LowerBound = new RSBound();
         RSBound UpperMinusLower = new RSBound();
         /* Generate where clause */
         String RecurseTargets = new String();
         for (java.util.Enumeration AllTargets = ColumnValues.elements();AllTargets.hasMoreElements();) {
            String ColumnLabel = (String)AllTargets.nextElement();
            if (RecurseTargets.length() < 2)
               RecurseTargets = new String(ColumnLabel+"='"+Combinations.getString(ColumnLabel)+"'");
            else
               RecurseTargets = new String(RecurseTargets+" AND "+ColumnLabel+"='"+Combinations.getString(ColumnLabel)+"'");
         }
         String SQLClassStatement = new String("SELECT SUM(_CARDINAL) FROM "+_table+" WHERE "+RecurseTargets+" AND "+ClassColumn+"='"+_class+"'");
         if (SDebug)
            System.err.println("RSRules.generateBounds: Sending SQL ["+SQLClassStatement+"].");
         java.sql.Statement CheckCombination = TheDatabase.createStatement();
         java.sql.ResultSet Checked = CheckCombination.executeQuery(SQLClassStatement);
         /* Check if this combination is any use. Skip if not. */
         if (Checked.next()) {
            LowerBound = new RSBound(RecurseTargets, Checked.getInt(1)); // Get good hits.
            SQLClassStatement = new String("SELECT SUM(_CARDINAL) FROM "+_table+" WHERE "+RecurseTargets+" AND "+ClassColumn+"<>'"+_class+"'");
            if (SDebug)
               System.err.println("RSRules.generateBounds: Sending SQL ["+SQLClassStatement+"].");
            Checked = CheckCombination.executeQuery(SQLClassStatement);
            if (Checked.next()) { // Not a deterministic rule!
               UpperMinusLower = new RSBound(RecurseTargets, Checked.getInt(1));
            }
         }
         /* Store results for this iteration if necessairy. */
         if (LowerBound.Coverage != 0)  {
            Bounds[0].addElement(LowerBound);
            Bounds[1].addElement(UpperMinusLower);
         } //if

      } //while
      return(Bounds);
   }//generateBounds
   /***
     * Generate an indiscernability set of the table which keeping track of table counts.
     * @returns The name of the temporairy synopsis table within the DBMS.
     * @param ColumnValuePairs Columns which form part of the indiscernability set.
     * @param _table The name of the table within the DBMS which represents the universe.
    ***/
   public String generateIndicernabilityFunction(java.util.Vector ColumnValuePairs, String _table) throws java.sql.SQLException {
       java.sql.Statement CheckTable = TheDatabase.createStatement();
       String OutputTable = this.generateTableName();
       String TargetList = new String("");
       for (java.util.Enumeration Targets = ColumnValuePairs.elements();Targets.hasMoreElements();)
          if (TargetList.length()<2)
             TargetList = new String((String)Targets.nextElement());
          else
             TargetList = new String(TargetList+", "+(String)Targets.nextElement());
       if (TargetList.length()>2)
          TargetList = new String(TargetList+", ");
       TargetList = new String(TargetList + ClassColumn);
       String SQLStatement = new String("SELECT COUNT("+ UNIQUEID+") AS _CARDINAL, "+TargetList+" INTO TEMP "+OutputTable+" FROM "+_table+" GROUP BY " +TargetList );
       if (SDebug)
          System.err.println("RSRules.getIndicernabilityFunction: Send SQL: ["+SQLStatement+"].");
       CheckTable.executeUpdate(SQLStatement);
       return(OutputTable);
   } // getIndicernabilityFunction()
   /***
     * Wrapper method to set the row idetifier column name.
    ***/
   public void setRowID(String _rowid) {
      UNIQUEID = _rowid;
   } // setRowID
   /***
     * Wrapper method to set the class column name.
    ***/
   public void setClass(String _class) {
      ClassColumn = _class;
   }//setClass
   /***
     *Wrapper method to set the universe table name.
    ***/
   public void setTable(String _table) {
      InformationTable = _table;
   }//setTable
   /***
     * Get all possible values of a column.
     * @returns A vector of possible values.
     * @param _Table The name of the table within the DBMS system.
     * @param _Column The name of the column within the table.
    ***/
    public java.util.Vector getValuesOfColumns(String _Table, String _Column) throws java.sql.SQLException {
       java.util.Vector ColumnNames = new java.util.Vector();
       java.sql.Statement CheckTable = TheDatabase.createStatement();
       java.sql.ResultSet TableResults = CheckTable.executeQuery("SELECT DISTINCT "+_Column+" FROM "+_Table);
       while (TableResults.next())
          ColumnNames.addElement(TableResults.getString(1));
       if (Debug)
          System.err.println("RSRules.getValuesOfColumns: Reports columns values as ["+ColumnNames.toString()+"].");
       return(ColumnNames);

    } // getValuesOfColumns
    /***
     * Return the number of rows within the table.
     * @returns The number of rows.
     * @param _Table The name of the table within the DBMS system.
    ***/
    public int getRowCount(String _Table) throws java.sql.SQLException {
       java.sql.Statement CheckTable = TheDatabase.createStatement();
       java.sql.ResultSet TableResults = CheckTable.executeQuery("SELECT COUNT("+UNIQUEID+") FROM "+_Table);
       TableResults.next();
       int MyRows = TableResults.getInt(1);
       if (Debug)
          System.err.println("RSRules.getRowCount: There are ["+MyRows+"] rows in table ["+_Table+"].");
       return(MyRows);
    } // getRowCount
    /***
     * Return the names of the columns within a table.
     * @returns The names of the column.
     * @param _Table The name of the table within the DBMS system.
    ***/
    public java.util.Vector getTableColumns(String _table) throws java.sql.SQLException {
       java.util.Vector ColumnNames = new java.util.Vector();
       java.sql.Statement CheckTable = TheDatabase.createStatement();
       java.sql.ResultSet TableResults = CheckTable.executeQuery("SELECT * FROM "+_table+" LIMIT 10");
       TableResults.next();
       java.sql.ResultSetMetaData TableData = TableResults.getMetaData();
       int ColumnCount = TableData.getColumnCount();
       for (int i=1;i<=ColumnCount;i++)
          ColumnNames.addElement(TableData.getColumnName(i));
       if (Debug)
          System.err.println("RSRules.getColumnName: Reports columns names as ["+ColumnNames.toString()+"].");
       return(ColumnNames);
    } // getTableColumns
    /***/
      * Generate a temporairy table name.
      * @returns The name of a table which does not already exist in the DBMs.
     ***/
    public String generateTableName() throws java.sql.SQLException {
       boolean LDebug = false;
       java.util.Random mynumbers = new java.util.Random();
       String _mytable = null;
       int anumber;
       if (LDebug)
          System.err.println("generateTableName: getting random number.");
       while (_mytable == null) {
          anumber = mynumbers.nextInt();
          if (anumber < 0)
             anumber = anumber * -1;
          _mytable = new String((new java.lang.Integer(anumber)).toString());
          _mytable= new String("TMP" + _mytable.trim());
          if (_mytable.length() >8)
             _mytable = _mytable.substring(0,7);
//          if (this.tableExists(_mytable))
//             _mytable = null;
       }
       if (LDebug)
          System.err.println("generateTableName: Returning ["+_mytable+"] as a table name.");
       return(_mytable);
    } //generateTableName
   /***
     * A replacement for the broken quality factor in RS1. 
     * @returns Shannon's disorder value for the given column.
     * @param ColumnTryout The name of the column for which we wish to have Shannon's number.
     * @param StatisticsTable The table name of the indiscernability set for this table.
     * @param ColumnsDone List of columns that are already being used as branches.
     * @param CurrentClass The current class values.
     * 
    ***/
   public double getShannon(String ColumnTryout, java.util.Vector ColumnsDone, String StatisticsTable, String CurrentClass) throws java.sql.SQLException {
      double ShannonResult = 0.0;
      java.sql.Statement CheckTable = TheDatabase.createStatement();
      java.sql.ResultSet TableResults = CheckTable.executeQuery("SELECT SUM(_CARDINAL) FROM "+StatisticsTable);
      TableResults.next();
      double RowsTotal = TableResults.getDouble(1);
      /* Generate a list of columns that we are concerned about. */
      String TargetList = new String(ColumnTryout);
      for (java.util.Enumeration Targets = ColumnsDone.elements();Targets.hasMoreElements();)
         TargetList = new String(TargetList+", "+(String)Targets.nextElement());
      String SQLStatement = new String("SELECT SUM(_CARDINAL) AS _CARDINAL,"+TargetList+" FROM "+StatisticsTable+" GROUP BY "+TargetList);
      if (SDebug)
         System.err.println("RSRules.getShannon: Sending SQL ["+SQLStatement+"].");   
      TableResults = CheckTable.executeQuery(SQLStatement);
      /***
        * We now have a listing of all possible branches that will need
        * to be taken for this list of columns along with the number of 
        * rows which are in each branch.
       ***/
       while (TableResults.next()) { 
          String WhereCondition = new String(ColumnTryout+"='"+TableResults.getString(ColumnTryout)+"'");
          for (java.util.Enumeration Targets = ColumnsDone.elements();Targets.hasMoreElements();) {
             String ThisColumn = (String)Targets.nextElement();
             WhereCondition = new String(WhereCondition+" AND "+ThisColumn+"='"+TableResults.getString(ThisColumn)+"'");
          } //for
          double BranchTotal = TableResults.getDouble("_CARDINAL");
          /***
            * Because in this case we are trying to generate a rule set for a specific
            * class, we can simplify Shannon's equation to a two class version: good class
            * bad class.
           ***/          
          java.sql.Statement ClassCheck = TheDatabase.createStatement();
          /* Count number of positive class instances. */
          String ClassSQL = new String("SELECT SUM(_CARDINAL) AS _CARDINAL FROM "+StatisticsTable+" WHERE "+WhereCondition+" AND "+ClassColumn+"='"+CurrentClass+"'");
          if (SDebug)
             System.err.println("RSRules.getShannon: Sending SQL ["+ClassSQL+"].");   
          java.sql.ResultSet ClassResult = ClassCheck.executeQuery(ClassSQL);
          ClassResult.next();
          double GoodClassesInBranch = ClassResult.getDouble(1);
          double GoodClassRatio = GoodClassesInBranch / BranchTotal;
          double GoodClassTerm = 0.0;
          if (GoodClassRatio != 0.0)
             GoodClassTerm = -GoodClassRatio * (java.lang.Math.log(GoodClassRatio)/java.lang.Math.log(2.0));
          if (Debug)
             System.err.println("RSRules.getShannon: ["+GoodClassesInBranch+"] good class  of ["+BranchTotal+"] rows. Ratio is ["+GoodClassRatio+"].");
          /* These are the wrong class */
          ClassSQL = new String("SELECT SUM(_CARDINAL) AS _CARDINAL FROM "+StatisticsTable+" WHERE "+WhereCondition+" AND "+ClassColumn+"<>'"+CurrentClass+"'");
          if (SDebug)
             System.err.println("RSRules.getShannon: Sending SQL ["+ClassSQL+"].");   
          ClassResult = ClassCheck.executeQuery(ClassSQL);
          ClassResult.next();
          double BadClassesInBranch = ClassResult.getDouble(1);
          double BadClassRatio = BadClassesInBranch / BranchTotal;
          double BadClassTerm =  0.0;
          if (BadClassRatio != 0.0)
             BadClassTerm = -BadClassRatio * (java.lang.Math.log(BadClassRatio)/java.lang.Math.log(2.0));
          if (Debug)
             System.err.println("RSRules.getShannon: ["+BadClassesInBranch+"] bad class  of ["+BranchTotal+"] rows. Ratio is ["+BadClassRatio+"].");
          /* Branch Total*/
          double TermTotal = BadClassTerm + GoodClassTerm;
          /*Shannon total */
          ShannonResult = ShannonResult + BranchTotal/RowsTotal*TermTotal;
       }//while
       if (Debug)
          System.err.println("RSRules.getShannon: Returning result:["+ShannonResult+"].");
      return(ShannonResult);
   } // getShannon
}//class


