import java.sql.*;
/***
  * An implementation of the RS1 Rough Sets algorythm. 
  * Class RoughSet is a clean implementation of the RoughSet method using the S.K.M. Wong
  * et al., Inductive Learning Algorythm RS1. (OBSOLETE)
  * 
  * Copyright 2000,2001,2002 Robert H. Warren 
  * 
  * RoughSet.java is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation; either version 2 of the License, or
  * (at your option) any later version.
  *
  * RoughSet.java is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
  * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  *
  * on-line information : http://www.gnu.org/licenses/gpl.txt
  * 
  *
  * @author Robert H. Warren
  * @author rhwarren@uwaterloo.ca
  * @version $Id: RoughSet.java,v 1.2 2002/10/11 02:34:55 warren Exp $
  * $Log: RoughSet.java,v $
  * Revision 1.2  2002/10/11 02:34:55  warren
  * Update contact information.
  * Mark as obsolete code.
  * Insert GPL blurb.
  *
  *
  * Revision 1.1  2000/09/12 02:03:45  warren
  * Initial revision
  *
  *
  ***/
public class RoughSet {
   /***
     * JDBC connection to the database in use.
    ***/
   Connection TheDatabase = null;
   /***
     * Vector of all table attributes names.
    ***/
   java.util.Vector AttributeNames = new java.util.Vector();
   /***
     * Keeps track of any temporary tables that were created during the processing.
    ***/
   java.util.Vector TableStack = new java.util.Vector();
   /***
     * Contains the column name of the row identifier.
    ***/
   String IDAttribute = null;
   /***
     * Contains the column name of the Class identifier.
    ***/
   String ClassAttribute = null;
   /***
     * Contains the value of the class we are currently looking for.
    ***/
   String ClassValue = new String("+");
   /***
     * Contains the values for the class that is considered to be positive.
    ***/
   String TableName = null;
   /***
     * @deprecated
     * Debugging is now local to the method.
     * Turns on (very) verbose debugging to System.err
    ***/
   boolean Debug = true;
   /***
     * Contains the name of the table that will contain Upper Bound rules.
    ***/
   String RuleTableH = new String("RULEH");
   /***
     * Contains the name of the table that will contain Lower Bound rules.
    ***/
   String RuleTableL = new String("RULEL");
/***
  * Main method is only used for testing. In the near furture, will put regression tests here.
  *
 ***/
   public static void main(String[] args) { 
      RoughSet mysets = null;
      try {
         System.out.println("RoughSet: Main: regression testing.");
         mysets = new RoughSet(new String("seq"), new String("object"),new String("item")); 
         mysets.cleanDB();	          
//         System.out.println("Upper boundairy rules: ");
//          for (java.util.Enumeration e = mysets.getRules(true).elements() ;e.hasMoreElements() ;) {            
//           System.out.println(e.nextElement());      
//   }          System.out.println("Lower boundairy rules: ");          
//          for(java.util.Enumeration e = mysets.getRules(false).elements() ;e.hasMoreElements() ;) {             
//            System.out.println(e.nextElement());      
//   }       
} catch (java.sql.SQLException e) {e.printStackTrace();}

       catch (java.lang.Exception e) {e.printStackTrace();mysets.tidyUp();}
       finally { mysets.tidyUp();
        }

   } //main
/***
  * Main constructor, and entry point, for the Rough Set engine. Will do limited sanity checking as well as
  * call the recursive rule generator method. Connection to the JDBC is hardcoded within this method.
  *
  * @param _id The column name of the unique row identifier.
  * @param _class The column name of the class identifier.
  * @param _table The table to be processed.
  * @exception java.lang.Exception Thrown in case of table error or parameter error.
  * @exception java.sql.SQLException Thrown in case of a problem with the jdbc driver.
  *
 ***/
   public RoughSet(String _id, String _class, String _table) throws java.sql.SQLException, java.lang.Exception {
      TableName = _table;
      ClassAttribute = _class;
      IDAttribute = _id;
      if (Debug)
         System.err.println("RoughSet: Connecting to database.");
      try {Class.forName("org.postgresql.Driver"); } catch (java.lang.ClassNotFoundException e) {e.printStackTrace();}
      TheDatabase =DriverManager.getConnection("jdbc:postgresql://switchboard.tzone.org/research","research", "mytools");
      Statement TestStatement = TheDatabase.createStatement();
      if (Debug)
         System.err.println("RoughSet: Checking if table ["+TableName+"] exists.");
      /*** Check that the table actually exists. ***/
      if (!tableExists(TableName)) {
         System.err.println("RoughSet: Could not find Table ["+TableName+"]"); 
         throw(new java.lang.Exception("Bad Table"));
      } //not table
      if (Debug)
         System.err.println("RoughSet: Loading column names.");
      ResultSet TestResults = TestStatement.executeQuery("SELECT * FROM "+TableName+" LIMIT 10");
      /*** Keep track of the column names. ***/
      ResultSetMetaData TableInfo = TestResults.getMetaData();
      int maxcolumns = TableInfo.getColumnCount();
      if (Debug)
         System.err.print("RoughSet: Loaded column:[");
      for (int i=1;i<=maxcolumns;i++) {
         AttributeNames.addElement(TableInfo.getColumnName(i).trim());
         if (Debug)
            System.err.print(TableInfo.getColumnName(i)+", ");
      } // for
      if (Debug)
         System.err.println("]");
      /*** Check that both class and id attributes are part of the columns. ***/
      if (Debug)
         System.err.println("RoughSet: Checking for Class Column ["+ClassAttribute+"].");
      if (!AttributeNames.contains(ClassAttribute.toLowerCase()))
         throw(new java.lang.Exception("BAD CLASS"));
      AttributeNames.removeElement(ClassAttribute);
      if (Debug)
         System.err.println("RoughSet: Checking for ID Column ["+IDAttribute+"].");
      if (!AttributeNames.contains(IDAttribute))
         throw(new java.lang.Exception("BAD ID"));
      AttributeNames.removeElement(IDAttribute);
      if (Debug)
         System.err.println("RoughSet: Cleaning up Upper Bound Rule Table ["+RuleTableH+"].");
      this.deleteTable(RuleTableH);
      TestStatement.executeUpdate("CREATE TABLE " +RuleTableH+ " () INHERITS ("+TableName+")");
      if (Debug)
         System.err.println("RoughSet: Cleaning up Upper Bound Rule Table ["+RuleTableL+"].");
      this.deleteTable(RuleTableL);
      TestStatement.executeUpdate("CREATE TABLE " +RuleTableL+ " () INHERITS ("+TableName+")");
      /****  Call recursive function ****/
/***********************************************************************/
      this.generateRules(TableName, null);
      this.generateResults();
   }//RoughSet
   /***
     * Recursive rule generation algorythm. The is the main processing loop.
     * 
     * @param _universe The name of the table or view that contains the current universe of rows.
     * @param _pivot Array of column names that are to be used as pivots during RS processing.
     * @exception java.sql.SQLException Thrown in case of a problem with the jdbc.
    ***/
   public void generateRules(String _universe, String[] _pivot) throws java.sql.SQLException {
      Statement rulestate = TheDatabase.createStatement();
      boolean LDebug = true;
      if (countRows(_universe) <=0) {
         if (LDebug)
            System.err.println("generateRules: No rows left in universe, exit.");
         return;
      }
      // Generate a view that will contain all class-positive example.
      String _class = this.createSubSet(_universe, str2ar(ClassAttribute), str2ar(ClassValue));
      if (LDebug)
         System.err.println("generateRules: Created table ["+_class+"] as position reference class.");
      /*** Initial case: Generate a new pivot. ***/
      if (_pivot == null) {
         if (LDebug)
	    System.err.println("generateRules: Pivo non-existent. Assume algorythm start.");
         java.util.Vector UpperBound = new java.util.Vector();
	 java.util.Vector LowerBound = new java.util.Vector();
	 double MaxAlpha = -1.0;
	 int MaxAlphaPosition = -1;
	 /*** Iterate for each attribute. ***/
	 for (java.util.Enumeration e = AttributeNames.elements();e.hasMoreElements();) {
	    String CurrentAttribute = (String)e.nextElement();
	    if (LDebug)
	       System.err.println("generateRules: Working on Attribute ["+CurrentAttribute+"].");
	    String _lowerbound = this.createSubSet(_universe,null,null);
	    if (LDebug)
	       System.err.println("generateRules: Using table ["+_lowerbound+"] to store Lower Bound.");
	    String _upperbound = this.createSubSet(_universe,null,null);
	    if (LDebug)
	       System.err.println("generateRules: Using table ["+_upperbound+"] to store Upper Bound.");
	    java.util.Vector AllValues = this.getAttributeValues(CurrentAttribute);
	    for (java.util.Enumeration ValueList = AllValues.elements();ValueList.hasMoreElements();) {
	       String CurrentValue = (String)ValueList.nextElement();
	       if (LDebug)
	          System.err.println("generateRules: Working on Attribute ["+CurrentAttribute+"] with value ["+CurrentValue+"].");
	       String SubSet = this.createSubSet(_universe, str2ar(CurrentAttribute),str2ar(CurrentValue));
	       this.generateBounds(str2ar(SubSet), _class, _upperbound, _lowerbound);	        
	    } // for all values
	    double thisalpha = this.generateAlpha(_universe, _upperbound, _lowerbound);
	    if (LDebug)
	       System.err.println("generateRules: Attribute ["+CurrentAttribute+"] has an alpha value of ["+thisalpha+"].");
            UpperBound.addElement(_upperbound);
            //Comment out rules...
            this.storeRules(_upperbound, str2ar(CurrentAttribute), ClassAttribute, true);
	    LowerBound.addElement(_lowerbound);
            this.storeRules(_lowerbound, str2ar(CurrentAttribute), ClassAttribute, false);
	    if (thisalpha > MaxAlpha) {
	       if (LDebug)
	          System.err.println("generateRules: This Alpha of ["+thisalpha+"] is better than best Alpha of ["+MaxAlpha+"].");
               MaxAlpha = thisalpha;
	       MaxAlphaPosition = (LowerBound.size()-1);
	    } // better alpha
	 } //for all attribute
	 if (LDebug)
	    System.err.println("generateRules: Best Attribute was ["+AttributeNames.elementAt(MaxAlphaPosition)+"] with an Alpha of ["+MaxAlpha+"], storing lower bound and recursing.");
	 this.storeRules((String)LowerBound.elementAt(MaxAlphaPosition), str2ar((String)AttributeNames.elementAt(MaxAlphaPosition)),ClassAttribute, true);
	 this.storeRules((String)UpperBound.elementAt(MaxAlphaPosition), str2ar((String)AttributeNames.elementAt(MaxAlphaPosition)),ClassAttribute, false);
         String[] thenewpivot = str2ar((String)AttributeNames.elementAt(MaxAlphaPosition));
         AttributeNames.removeElementAt(MaxAlphaPosition);
         generateRules(_universe, thenewpivot);      
      } else {
	 /***
	  * In this case, we begin by getting all possible values for each pivot and try each combination.
	  * Each time a set is generated and bounds computed. Things are more complicated because of the
          * multiple sets needed.
	  *
	  * Should change all of this by generating a new function called find pivot.
	 ***/
         if (LDebug)
	    System.err.println("generateRules: Pivoting on at least ["+_pivot[0]+"].");
	 //get all pivot values
	 java.util.Vector[] PivotValues = new java.util.Vector[_pivot.length+1];
	 for (int i=0;i<_pivot.length;i++) {
	    if (LDebug)
	       System.err.println("generateRules: Getting possible values for pivot ["+_pivot[i]+"].");
	    PivotValues[i] = this.getAttributeValues(_pivot[i]);
	 }//for all pivots

	 // Do same as simple case, but for each pivot.

	 java.util.Vector TestSets = this.generateComplexSets(_pivot, PivotValues);
            java.util.Vector UpperBound = new java.util.Vector();
            java.util.Vector LowerBound = new java.util.Vector();
	    double MaxAlpha = -1.0;
	    int MaxAlphaPosition = -1;
         for (java.util.Enumeration OtherPivots = TestSets.elements();OtherPivots.hasMoreElements();) {
	    String ThePivot = (String)OtherPivots.nextElement();
            if (LDebug)
	       System.err.println("generateRules: The Pivot constraits are ["+ThePivot+"].");
	    /*** Iterate for each attribute. ***/
	    for (java.util.Enumeration e = AttributeNames.elements();e.hasMoreElements();) {
	       String CurrentAttribute = (String)e.nextElement();
	       boolean okay = true;
	       for (int i=0;i<_pivot.length;i++)
	          if (_pivot[i].equals(CurrentAttribute))
	             okay=false;
	       if (okay) {	  
      	          if (LDebug)
	             System.err.println("generateRules: Working on Attribute ["+CurrentAttribute+"].");
	             String _lowerbound = this.createSubSet(_universe,null,null);
	             if (LDebug)
	                System.err.println("generateRules: Using table ["+_lowerbound+"] to store Lower Bound.");
	             String _upperbound = this.createSubSet(_universe,null,null);
	             if (LDebug)
	                System.err.println("generateRules: Using table ["+_upperbound+"] to store Upper Bound.");
	             java.util.Vector AllValues = this.getAttributeValues(CurrentAttribute);
	             for (java.util.Enumeration ValueList = AllValues.elements();ValueList.hasMoreElements();) {
	                String CurrentValue = (String)ValueList.nextElement();
	                if (LDebug)
	                   System.err.println("generateRules: Working on Attribute ["+CurrentAttribute+"] with value ["+CurrentValue+"].");
		        String trytable = this.generateTableName();
		        Statement astatement = TheDatabase.createStatement();
		        String SQLSOUP = new String("CREATE TABLE " +trytable + " () INHERITS ("+_universe+")");
		        if (LDebug)
		           System.err.println("generateRules: Create table with SQLSOUP ["+SQLSOUP+"].");
		        int done = astatement.executeUpdate(SQLSOUP);
		  
	   	        SQLSOUP = new String("INSERT INTO " + trytable + " SELECT * from "+_universe+" WHERE "
		        +CurrentAttribute+"='"+CurrentValue+"' AND " + ThePivot);
		        if (LDebug)
		           System.err.println("generateRules: Populate table with SQL SOUP ["+SQLSOUP+"].");
 		        done = astatement.executeUpdate(SQLSOUP);
			if (LDebug)
			   System.err.println("generateRules: Inserted ["+done+"] rows.");  
	                this.generateBounds(str2ar(trytable), _class, _upperbound, _lowerbound);	        
	             } // for all values

                  double thisalpha = this.generateAlpha(_universe, _upperbound, _lowerbound);
	          if (LDebug)
	             System.err.println("generateRules: Attribute ["+CurrentAttribute+"] has an alpha value of ["+thisalpha+"].");
                  UpperBound.addElement(_upperbound);
                  // Comment out rues
                  this.storeRules(_upperbound, str2ar(CurrentAttribute), ClassAttribute, true);
	          LowerBound.addElement(_lowerbound);
                  this.storeRules(_lowerbound, str2ar(CurrentAttribute), ClassAttribute, false);
	          if (thisalpha > MaxAlpha) {
	             if (LDebug)
	                System.err.println("generateRules: This Alpha of ["+thisalpha+"] is better than best Alpha of ["+MaxAlpha+"].");
                     MaxAlpha = thisalpha;
	             MaxAlphaPosition = (LowerBound.size()-1);
	          } // better alpha
               }
	    } //for all attribute
         }//for all pivots
	 if (LDebug)
	    System.err.println("generateRules: Best Attribute was ["+AttributeNames.elementAt(MaxAlphaPosition)+"] with an Alpha of ["+MaxAlpha+"], storing lower bound and recursing.");
         String[] newpivot = new String[(_pivot.length+1)];
	 for (int i=0;i<_pivot.length;i++)
	    newpivot[i]=_pivot[i];
	 newpivot[(_pivot.length)] = (String)AttributeNames.elementAt(MaxAlphaPosition);   
	 this.storeRules((String)LowerBound.elementAt(MaxAlphaPosition), newpivot,ClassAttribute, true);
	 this.storeRules((String)UpperBound.elementAt(MaxAlphaPosition), newpivot,ClassAttribute, false);

         if (LDebug)
            System.err.println("generateRules: New pivot is ["+newpivot[(_pivot.length)]+"].");
         AttributeNames.removeElementAt(MaxAlphaPosition);
         _universe = reduceUniverse(_universe, (String)LowerBound.elementAt(MaxAlphaPosition), (String)UpperBound.elementAt(MaxAlphaPosition));
         generateRules(_universe, newpivot);
	 System.out.println("*** Finished processing ***");
      } //multiple aatribute land...   
   }//generateRules
   /***
     * Create a new universe using a view. 
     *
     * @param _olduniverse The name of the table that was the old universe.
     * @param _lowerbound The name of the table that is the Lower Bound.
     * @param _upperbound The name of the table that is the Upper Bound.     
     * @return The name of the table that is the new universe.
    ***/
   public String reduceUniverse(String _olduniverse, String _lowerbound, String _upperbound) throws java.sql.SQLException {
      boolean LDebug = false;
      if (LDebug)
         System.err.println("reduceUniverse: Will generate a new universe from ["
         +_olduniverse+"] with upperbound ["+_upperbound+"] and lowerbound ["+_lowerbound+"].");
      // Universe - Upperbound
      String temptable = this.generateTableName();
      if (LDebug)
         System.err.println("reduceUniverse: First temp table is ["+temptable+"].");

      this.substractSet(_upperbound, _olduniverse, temptable);
      // (Universe - UpperBound) union (LowerBound)
      String temptable2 = this.createSubSet(_olduniverse, null, null);
      if (LDebug)
         System.err.println("reduceUniverse: First temp table is ["+temptable2+"].");

      this.addToTable(temptable, temptable2);
      this.addToTable(_olduniverse, temptable2);
      // Universe - (Universe - UpperBound) union (LowerBound);

      String NewUniverse = this.generateTableName();
      if (LDebug)
         System.err.println("reduceUniverse: First temp table is ["+NewUniverse+"].");
      this.substractSet(temptable2, _olduniverse, NewUniverse);
      //return
      return(NewUniverse);
   } // reduce universe.

   /***
     * A recursive SQL condition generator used internally to generate the WHERE clause when multiple
     * attributes and/or values occur. Actually simplifies query generation.
     *
     * @param pivot An array of attribute names.
     * @param Values An array that map a Vector of values to each attribute.
     * @returns A Vector of possible WHERE clauses.
    ***/
   public java.util.Vector generateComplexSets(String[] pivot, java.util.Vector[] Values) {
      boolean LDebug = false;
      if (LDebug)
         System.err.println("genereateComplexSets: Entering methods.");
      if (LDebug)
         System.err.println("genereateComplexSets: There are ["+pivot.length+"] pivots and ["+Values.length+"] sets of values.");
      if (pivot.length==1) {
      if (LDebug)
         System.err.println("generateComplexSets: Pivot is ["+pivot[0]+"].");
      if (LDebug)
         System.err.println("genereateComplexSets: Entering simplex case with values ["+Values[0].toString()+"].");
         java.util.Vector condition = new java.util.Vector();
         
         for (java.util.Enumeration e = Values[0].elements();e.hasMoreElements();) {
	    String mystring = new String(" " + pivot[0] + "='" + (String)e.nextElement()+"' ");
	    condition.addElement(mystring);
	 }
	 return(condition);
      } else
      {
         if (LDebug)
            System.err.println("genereateComplexSets: Entering complex case.");
         String[] newpivot = new String[(pivot.length-1)];
	 java.util.Vector[] newvector = new java.util.Vector[pivot.length];
	 for (int i=1;i<=(pivot.length-1);i++) {
            if (LDebug)  
               System.err.println("generateComplexSets: Copying pivot ["+pivot[i]+"] with values ["+Values[i]+"].");
	    newpivot[i-1] = pivot[i];
	    newvector[i-1]=Values[i];
	 }//for
	 java.util.Vector returning = this.generateComplexSets(newpivot, newvector);

	 java.util.Vector condition = new java.util.Vector();
         for (java.util.Enumeration e = Values[0].elements();e.hasMoreElements();) {
            String elementvalue = (String)e.nextElement();
            for (java.util.Enumeration y = returning.elements();y.hasMoreElements();) {
	       String mystring = new String(" " + pivot[0] + "='" + elementvalue +"' AND " + (String)y.nextElement());
	       condition.addElement(mystring);
            }
	 }//for
	 return(condition);
      }//if
   }//complex rules
   /***
     * Generates a subset table based on a list of attribute/value pairs. If the values of <CODE>ValueList</CODE>
     * and <CODE>AttributeList</CODE> are null, then a blank table will be generated with the same structure as
     * <CODE>_universe</CODE>.
     *
     * @param _universe The set that the sub set will be generated from.
     * @param AttributeList An array of attributes matched to<CODE>ValueList</CODE>.
     * @param ValueList An array of values matched to<CODE>AttributeList</CODE>.
     * @return The table name of the subset.
     *
    ***/
   public String createSubSet(String _universe, String[] AttributeList, String[] ValueList) throws java.sql.SQLException {
      boolean LDebug = true;
      Statement thestatement = TheDatabase.createStatement();
      String outputtable = this.generateTableName();
      if (LDebug)
         System.err.println("createSubSet: Will write out a subset to table ["+outputtable+"].");
      String SQLSOUP = new String("CREATE TABLE "+outputtable+" () INHERITS ("+_universe+")");
      if (LDebug)
         System.err.println("createSubSet: Creating table with SQL SOUP ["+SQLSOUP+"].");
      int creation = thestatement.executeUpdate(SQLSOUP);
      if (AttributeList != null) {
         if (LDebug)
	    System.err.println("createSubSet: Will need to check for ["+AttributeList.length+"] attributes.");
         boolean first = true;
	 SQLSOUP = new String();
	 for (int i=0; i<AttributeList.length;i++) {
	    if (!first) {
	       SQLSOUP = new String(SQLSOUP + " AND ");	       
	    } else
	       first = false;
	    SQLSOUP = new String(SQLSOUP + AttributeList[i] + "='" + ValueList[i] + "'");
	 }//for
	 SQLSOUP = new String("INSERT INTO " + outputtable + " SELECT * FROM " + _universe + " WHERE " + SQLSOUP);
	 if (LDebug)
	    System.err.println("createSubSet: Will insert rows using SQL SOUP ["+SQLSOUP+"].");
	 creation = thestatement.executeUpdate(SQLSOUP);
	 if (LDebug)
	    System.err.println("createSubSet: Inserted ["+creation+"] rows into table.");
      } // need to populate
      return(outputtable);
   } //createSubSet

   /***
     * Stores the relevant upperbound/lowerbound rules within the appropriate tables, as
     * defined by the class Strings <CODE>RuleTableL</CODE> and <CODE>RuleTableH</CODE>.
     * Upper Bound rules are recursively inserted within the Lower Bound. 
     *
     * @param _table The table containing the actual rules.
     * @param _columns An array containing the current Attributes that are 
     * actually part of the rules.
     * @param _class The current value for the class that this ruleset selects for.
     * @param lowerbound Set to true if the current ruleset is part of the lower bound ruleset.
    ***/
   public void storeRules(String _table, String[] _columns, String _class, boolean lowerbound) throws java.sql.SQLException {
      boolean LDebug = true;
      Statement mystate = TheDatabase.createStatement();
      Statement output = TheDatabase.createStatement();
      if (LDebug)
         System.err.print("storeRules: Storing ");
      if ((LDebug) && (lowerbound))
         System.err.print("lower bound ");       
      if ((LDebug) && (!lowerbound))
         System.err.print("upper bound ");
      String OutputTable;
      if (lowerbound)
         OutputTable = RuleTableL;
      else
         OutputTable = RuleTableH;
      if (LDebug)
         System.err.println("rule(s) over ["+_columns.length+"] columns from table ["+_table+"] to table ["+OutputTable+"].");
      /*** Make sure to get the proper combinations from the table ***/
      boolean first = true;
      String Selection = new String();
      for (int i=0;i<_columns.length;i++) {
         if (!first) {
	    Selection = new String(Selection + ", ");
         } else {
            first = false;
         }
         Selection = new String(Selection + " " + _columns[i]);
      }//for
      Selection = new String(Selection +", " + _class);	 

      String SQLSOUP = new String("SELECT DISTINCT "+Selection+"  FROM "+_table);
      if (LDebug)
         System.err.println("storeRules: Will send ["+SQLSOUP+"] as SQL Soup.");
      ResultSet TheRules = mystate.executeQuery(SQLSOUP);
      while (TheRules.next()) {
         String Rule = new String();
	 boolean firsta = true;
         for (int i=0;i<_columns.length;i++) {
            if (firsta)
	       firsta = false;
	    else
	       Rule = new String(Rule+ ", ");
            Rule = new String(Rule + " '" + TheRules.getString(_columns[i])+"'");
         }//for
         Rule = new String(Rule +", '" + TheRules.getString(_class)+"'");	 
         SQLSOUP = new String("INSERT INTO " +OutputTable + " ("+Selection+") VALUES ("+Rule+")");
	 if (LDebug)
	    System.err.println("storeRules: Will send ["+SQLSOUP+"] as SQL Soup.");
	 int sent = output.executeUpdate(SQLSOUP);
         if (LDebug)
            System.err.println("storeRules: executeUpdate reported ["+sent+"] row(s) inserted.");
         	 
      }
      if (lowerbound) {
         if (LDebug)
            System.err.println("storeRules: Recursing to store the lowerbound rules from a lowerbound rule.");
        this.storeRules(_table, _columns, _class, false);
      }
   }//storeRules

   /***
     * Generate the alpha measure of a Upperbound and Lowerbound ruleset on a
     * universe. Note that the result is 1 if no deterministic (Lower Bound
     * is empty) rules exists. This can leads to rule generation problems.
     * This is a flaw with the RS1 algorythm, a backup value for alpha should
     * be used also.
     * 
     * 
     * @param _universe The table name pointing to the current universe.
     * @param _upperbound The table name where the upper bound rules are stored.
     * @param _lowerbound The table name where the lower bound rules are stored.
     * @exception java.sql.SQLException Thrown if some problem occurs with the jdbc.
     * @returns The value of alpha for this ruleset. 
    ***/
   public double generateAlpha(String _universe, String _upperbound, String _lowerbound) throws java.sql.SQLException {
      boolean LDebug = true;
      double uppercount = (double)this.countRows(_upperbound);
      if (LDebug)
         System.err.println("generateAlpha: Got Upper Bound of ["+uppercount+"].");
      double lowercount = (double)this.countRows(_lowerbound);
      if (LDebug)
         System.err.println("generateAlpha: Got Lower Bound of ["+lowercount+"].");
      double unicount = (double)this.countRows(_universe);
      if (LDebug)
         System.err.println("generateAlpha: Got Universe count of ["+unicount+"].");
      double alpha = ((double)1 - (uppercount - lowercount)/unicount);
      if (LDebug)
         System.err.println("generateAlpha: Got Alpha of ["+alpha+"].");
      return(alpha);
   } //generateAlpha
   /***
     * Insert records from one table to another. This assumes that both
     * tables have the same fields. 
     * @param _from The table name of the source table.
     * @param _target The table name of the target table.
     * @exception java.sql.SQLException If anything should go wrong with the jdbc, 
     * including incompatible tables.
    ***/
   public void addToTable(String _from, String _target) throws java.sql.SQLException {
      boolean LDebug = false;
      Statement copyer = TheDatabase.createStatement();
      if (LDebug)
         System.err.println("addToTable: Copying record(s) from table ["+_from+"] to table ["+_target+"].");
      String SQLSOUP = new String("INSERT INTO "+_target+" SELECT * FROM "+_from);
      if (LDebug)
         System.err.println("addToTable: Will send ["+SQLSOUP+"] as SQL soup.");
      int copied = copyer.executeUpdate(SQLSOUP);
      if (LDebug)
         System.err.println("addToTable: Copied ["+copied+"] record(s).");   
   }//addtotable
   
   /***
     * Generate low and upper rulesets for each subset. Given an array of sets, will 
     * seperate them into one set of upperbound rules and one set of lowerbound rules. 
     * 
     * @param _subset An array of sets to be processed.
     * @param _upperbound The table name where upperbound rulesets will be stored.
     * @param _lowerbound The table name where lowerbound rulesets will be stored.
     * @param _class The table name of the positive classes we are looking for.
     * @exception java.sql.SQLException Thrown if a problem occurs with the database.
    ***/
   public void generateBounds(String[] _subset, String _class, String _upperbound, String _lowerbound) throws java.sql.SQLException {
      boolean LDebug = true;
      if (LDebug)
         System.err.println("generateBound: Will process the upper/lower bounds for ["+_subset.length+"] subset(s).");
      for (int i=0;i<_subset.length;i++) {
         if (LDebug)
	    System.err.println("generateBound: Processing subset ["+_subset[i]+"].");
         if (this.isIncluded(_subset[i], _class)) {
	    if (LDebug)
	       System.err.println("generateBound: Subset ["+_subset[i]+"] is included within class ["+_class+"].");
	    this.addToTable(_subset[i], _lowerbound);
	    this.addToTable(_subset[i], _upperbound);
	 } else if (!this.intersectionEmpty(_subset[i], _class)) {
	    if (LDebug)
	       System.err.println("generateBound: Subset ["+_subset[i]+"] intersects with class ["+_class+"].");
	    this.addToTable(_subset[i], _upperbound);
	 } else if (LDebug)
	    System.err.println("generateBound: Nothing to do with Subset ["+_subset[i]+"].");
      }//for
   }//generatebound
   /***
     * Checks whether there exists elements within the intersection of both tables.
     * @param _table1 The name of the first table to be intersected.
     * @param _table2 The name of the second table to the intersected.
     * @returns true if the intersection is empty, false if at least one element was present.
     * @exception java.sql.SQLException Thrown if a problem occurs with the jdbc, of if one of the
     * tables is missing.
    ***/
   public boolean intersectionEmpty(String _table1, String _table2) throws java.sql.SQLException {
      boolean LDebug = false;
      Statement mystatement = TheDatabase.createStatement();
      String SQLSOUP = new String("SELECT "+IDAttribute+" FROM "+_table1+" WHERE "
      +IDAttribute+" IN (SELECT "+ IDAttribute +" FROM "+ _table2 +") LIMIT 1");
      if (LDebug)
         System.err.println("intersectionEmpty: Trying to see if there is an intersection between ["+_table1+"] and ["+_table2+"].");
      if (LDebug)
         System.err.println("intersectionEmpty: Sending SQL Soup ["+SQLSOUP+"].");      
      ResultSet intersection = mystatement.executeQuery(SQLSOUP);
      if (intersection.next()) {
         if (LDebug)
            System.err.println("intersectionEmpty: Data present in the intersection.");
         return(false);
      } else {
         if (LDebug)
            System.err.println("intersectionEmpty: No data present in the intersection.");
         return(true);
      }      
   }// intersection empty.
   /***
     * Returns true if <CODE>table1</CODE> is included within <CODE>_table2</CODE>.
     * @param _table1 Is the set that is included.
     * @param _table2 Is the set that is included from.
     * @returns true if _table1 is included within _table2.
     * @exception java.sql.SQLException Thrown if anything goes wrong with the jdbc, including
     * missing table.
    ***/
   public boolean isIncluded(String _table1, String _table2) throws java.sql.SQLException {
      boolean LDebug = false;
      boolean answer = false;
      Statement Astatement = TheDatabase.createStatement();
      if (LDebug)
         System.err.println("isIncluded: Trying to find out if ["+_table1
         +"] is included within ["+_table2+"].");
      String tempview = this.generateTableName();
      if (LDebug)
         System.err.println("isIncluded: Using temporairy view ["+tempview+"].");
      if (this.substractSet(_table2, _table1, tempview) > 0)
         answer = false;
      else
         answer = true;
      if (LDebug)
         System.err.println("isIncluded: Deleting temporairy view ["+tempview+"].");
      this.deleteTable(tempview);
      if (LDebug)
         System.err.println("isIncluded: Returning ["+answer+"].");
      return(answer);      
   } //isIncluded
   /***
     * Count the number of unique rows in the table/view.
     * @param _table The table or view to count the rows from.
     * @return The total number of rows. -1 is something's wrong.
     * @exception java.sql.SQLException If anything should happen to the jdbc.
    ***/
   public int countRows(String _table) throws java.sql.SQLException {
      boolean LDebug = false;
      int totalcount;
      Statement counting = TheDatabase.createStatement();
      String SQLSOUP = new String("SELECT DISTINCT COUNT("+IDAttribute+") from " + _table);
      if (LDebug)
         System.err.println("countRows: Will count the total number of rows in table ["+_table+"].");
      if (LDebug)
         System.err.println("countRows: Sending the SQL Soup ["+SQLSOUP+"].");
      ResultSet counts = counting.executeQuery(SQLSOUP);
      if (counts.next()) 
         totalcount = counts.getInt(1);
      else
         totalcount = -1;
      if (LDebug)    
         System.err.println("The total number of rows in table ["+_table+"] is "+totalcount+".");
      return(totalcount);
   }//countRows
   /***
     * Delete the table/view from the database. Does this by first deleting the view,
     * or if an exeception occurs, then just try as a table. This will never
     * throw an exception, as this is meant as a fail-safe method. 
     * @param _table The name of the view/table to delete. 
     * @return true if the delete was succesfull, false if a problem occurred.
    ***/
   public boolean deleteTable(String _table) {
      boolean LDebug = false;
      try {
      Statement mystatement = TheDatabase.createStatement();
      boolean done = false;
      if (LDebug)
         System.err.println("deleteTable: Trying to delete view/table ["+_table+"].");
      try {
         if (LDebug)
            System.err.println("deleteTable: Trying to delete as a view.");
         String SQLSOUP = new String("DROP VIEW "+ _table);
         if (LDebug)
            System.err.println("deleteTable: Sending SQLSOUP ["+SQLSOUP+"].");
         int trash = mystatement.executeUpdate(SQLSOUP);
         done = true;
      } catch (java.sql.SQLException e) {
         if (LDebug) 
            System.err.println("deleteTable: Tried to delete ["+_table+"] as a view and got exception ["+e.getMessage()+"].");
         done = false;
      } // try as view
      if (!done) {
         try {
            if (LDebug)
               System.err.println("deleteTable: trying to delete as a table.");
            String SQLSOUP = new String("DROP TABLE "+ _table);
            if (LDebug)
               System.err.println("deleteTable: Sending SQLSOUP ["+SQLSOUP+"].");
            int trash = mystatement.executeUpdate(SQLSOUP);
            done = true;
         } catch (java.sql.SQLException e) { 
            if (LDebug)
               System.err.println("deleteTable: Tried to delete ["+_table+"] as a table and got exception ["+e.getMessage()+"].");
            done = false;
         } //try as table
      } //!done
      return(done);
      } catch (java.sql.SQLException e) {}
      return(false);
   } //deleteTable
   /***
     *  Returns the substraction of one table from another. 
     *
     * @param _tablefrom Is the table that is substracted from.
     * @param _tablesub Is the table that is substracted.
     * @param _target Is the target view that is to be created.
     * @return The number of rows actually created 
     * @exception java.sql.SQLException if anything bad happends with the jdbc.
    ***/
   public int substractSet(String _tablesub, String _tablefrom, String _target) throws java.sql.SQLException {
      boolean LDebug = false;
      Statement Astatement = TheDatabase.createStatement();
      if (LDebug)
         System.err.println("substractSet: Subtracting set ["+_tablesub+"] from set ["+_tablefrom+"] and putting the result in set ["+_target+"].");
      String SQLSOUP = new String("CREATE VIEW "+_target+" AS SELECT * FROM "+_tablefrom+" WHERE "+IDAttribute+" NOT IN (SELECT "+IDAttribute+" FROM "+_tablesub+")");
      if (LDebug)
         System.err.println("substractSet: Sending SQLSOUP ["+SQLSOUP+"].");
      int counter;
      counter = Astatement.executeUpdate(SQLSOUP);
      if (LDebug)
         System.err.println("substractSet: executeUpdate returned value "+counter+".");
      counter = this.countRows(_target);
      if (LDebug)
         System.err.println("substractSet: View has "+counter+" row(s).");
      return(counter);
   }
   /***
     * Create a view that is the intersection of both tables. Excluding the ClassAttribute
     * and IDAttribute in the joins.
     * 
     * @param _table1 The first table to be intersected.
     * @param _table2 The second table to be intersected.    
     * @param _target The name of the table to be generated.
     * @return The number of rows that make up the interesection.
     * @exception java.sql.SQLException In case something bad happens with the jdbc.
     * 
    ***/
    public int makeIntersection(String _table1, String _table2, String _target) throws java.sql.SQLException {
       boolean LDebug = false;
       Statement Astatement = TheDatabase.createStatement();
       if (LDebug)
          System.err.println("makeIntersection: About to generate view ["+_target
          + "] from table ["+_table1+"] and table ["+_table2+"].");
       String SQLSOUP = new String("CREATE VIEW "+_target+" AS SELECT * from "
       +_table1+" WHERE "+IDAttribute+" in (SELECT "+IDAttribute+" from "+_table2+")");
       if (LDebug)
          System.err.println("makeIntersection: Sending SQLSOUP as ["+SQLSOUP+"]");
       int viewcount;
       viewcount = Astatement.executeUpdate(SQLSOUP);
       if (LDebug)
          System.err.println("makeIntersection: Update returned a result of "+viewcount+".");
       /***************************/
       SQLSOUP = new String("SELECT DISTINCT COUNT("+IDAttribute+") FROM "+_target);       
       if (LDebug)
          System.err.println("makeIntersection: Sending SQLSOUP as ["+SQLSOUP+"]");
       ResultSet HowMany = Astatement.executeQuery(SQLSOUP);
       if (HowMany.next())
          viewcount = HowMany.getInt(1);
       else
          viewcount = -1;
       return(viewcount);

    } //makeIntersection

   /***
     * Returns a vector with all possible values for the column name given.
     * 
     * @param _attribute The name of the attribute (column) to be used.
     * @return A vector will all of the possible values.
     * @exception java.sql.SQLException Thrown if a problem occurs with the jdbc driver.
     *
    ***/
   public java.util.Vector getAttributeValues(String _attribute) throws java.sql.SQLException {
      boolean LDebug = false;
      java.util.Vector _myvector = new java.util.Vector();
      Statement Astatement = TheDatabase.createStatement();
      String SQLSoup = new String("SELECT DISTINCT " + _attribute + " FROM " + TableName);
      if (LDebug)
         System.err.println("getAttribute: Sending SQLSOUP as ["+SQLSoup+"]");
      ResultSet myResults = Astatement.executeQuery(SQLSoup);
      if (LDebug)
         System.err.print("[");
      while (myResults.next()) {
         String _myvalue = myResults.getString(1).trim();
         if (_myvalue.length() > 0) {
            _myvector.addElement(_myvalue);
            if (LDebug)
               System.err.print(_myvalue+", ");
         }//if
         
      } //while
      if (LDebug)
         System.err.println("]");
      return(_myvector);
   }//getSttributeValues
   /***
     * Checks wheter a table exists or not within the database.
     * @param _table The name of the table.
     * @return true if the table exists, false if it does now.
     * @throws java.sql.SQLException Thrown if a problem occurs within
     * the jdbc, but not if the table is non-existent.
    ***/
    public boolean tableExists(String _table) throws java.sql.SQLException {
       boolean LDebug = false;
       Statement mystatement = TheDatabase.createStatement();
       try {
          ResultSet theresults = mystatement.executeQuery("SELECT * FROM "+_table + " LIMIT 1");
          if (theresults.next())
             return(true);
          else
             return(true);
       } catch (java.sql.SQLException e) {
          if (LDebug)
             System.err.println("Couldn't find table [" + _table + "]");
          if (LDebug)
             System.err.println("SQL Message " + e.getMessage());
          return(false);
       } //catch
    } //tableExists
    /***
      * Generate a random table name that does not exist within the database. Name format
      * is TMPXXXXX where XXXXX is a number between 0 and 99999. 
      * @return A 8 character long random table name.
      * @exception java.sql.SQLException Thrown if a problem occurs with the JDBC.
     ***/ 
    public String generateTableName() throws java.sql.SQLException {
       boolean LDebug = false;
       java.util.Random mynumbers = new java.util.Random();
       String _mytable = null;
       int anumber;
       if (LDebug)
          System.err.println("generateTableName: getting random number.");
       while (_mytable == null) {
          anumber = mynumbers.nextInt();
          if (anumber < 0)
             anumber = anumber * -1;
          _mytable = new String((new java.lang.Integer(anumber)).toString());
          _mytable= new String("TMP" + _mytable.trim());
          if (_mytable.length() >8) 
             _mytable = _mytable.substring(0,7);
          if (this.tableExists(_mytable))
             _mytable = null;
       }
       if (LDebug)
          System.err.println("generateTableName: Returning ["+_mytable+"] as a table name.");
       TableStack.addElement(_mytable);
       return(_mytable);
    } //generateTableName
   /***
     * This methods deletes all of the temporary tables/views that were created during
     * the current RoughSet processing. Will always succeed, even with jdbc trouble.
    ***/
   public void tidyUp() {
      boolean LDebug = false;
      for (java.util.Enumeration temptables = TableStack.elements();temptables.hasMoreElements();) {
         String deleteme = (String)temptables.nextElement(); 
         if (LDebug)
	    System.err.println("tidyUp: Removing table/view ["+deleteme+"].");
	 this.deleteTable(deleteme);
      }//for
   }//tidyup
   /***
     * A short stand alone method to clean up the database from all TMPXXXXX format table. Intended as a 
     * way to clean up after a crash.
    ***/
   public void cleanDB() {
      boolean LDebug = true;
      int counter =0;
      String _mytable;
       while (counter <=9999) {
          _mytable = new String((new java.lang.Integer(counter)).toString());
          _mytable= new String("TMP" + _mytable.trim());
          if (_mytable.length() >8) 
             _mytable = _mytable.substring(0,7);
          if (LDebug)
             System.err.println("cleanDB: Delete table ["+_mytable+"].");
	  this.deleteTable(_mytable);
          counter++;
       } //while

   } // cleanDB
   /***
     * A string to string array converter. Used internally to get around java 
     * compiler problems.
     * @returns An array of one string.
     * @param _thestring The String to be converted.
    ***/
   public String[] str2ar(String _thestring) {
      //if (Debug)
      //   System.err.println("str2ar: Trying to convert ["+_thestring+"].");
      String[] myarray = new String[1];
      myarray[0] = new String(_thestring);
      return(myarray);
   } // str2ar
   /***
     * Reads upperbound/lowerbound rule tables, adds confidence and support rows.
     * and tally them up. Actually wraps <CODE>generateResultsTable()</CODE>.
     * @exception java.sql.SQLException Thrown if the jdbc has a problem.
    ***/
   public void generateResults() throws java.sql.SQLException  {
      this.generateResultsTable(RuleTableH);
      this.generateResultsTable(RuleTableL);           
   } //generateResults
   /***
     * For each rule within the upper/lower ruleset, generate the needed
     * statitics by adding the <CODE> POSITIVE, NEGATIVE </CODE> and <CODE> COVERAGE </CODE>
     * columns for each rule. 
     * 
     * @param _table The table to process.
     * @expection java.sql.SQLException Thrown if the jdbc has a problem.
    ***/
   private void generateResultsTable(String _table) throws java.sql.SQLException {
      boolean LDebug = false;
      java.sql.Statement mystate = TheDatabase.createStatement();
      String temptable = this.createSubSet(TableName,null,null);
      String SQLSOUP = new String("ALTER TABLE "+temptable+" ADD COLUMN POSITIVE int");
      if (LDebug)
         System.err.println("generateResults: Modifying table ["+temptable+"] with SQL soup ["+SQLSOUP+"].");
      mystate.executeUpdate(SQLSOUP);
      SQLSOUP = new String("ALTER TABLE "+temptable+" ADD COLUMN NEGATIVE int");
      if (LDebug)
         System.err.println("generateResults: Modifying table ["+temptable+"] with SQL soup ["+SQLSOUP+"].");
      mystate.executeUpdate(SQLSOUP);
      SQLSOUP = new String("ALTER TABLE "+temptable+" ADD COLUMN COVERAGE int");
      if (LDebug)
         System.err.println("generateResults: Modifying table ["+temptable+"] with SQL soup ["+SQLSOUP+"].");
      mystate.executeUpdate(SQLSOUP);
      

      /********************************************/
      if (LDebug)
         System.err.println("generateResults: Processing table ["+_table+"].");
      SQLSOUP = new String("Select DISTINCT * from " + _table + " WHERE "+ClassAttribute+"='"+ClassValue+"'");
      java.sql.ResultSet Rules = mystate.executeQuery(SQLSOUP);
      java.sql.ResultSetMetaData RulesMeta = Rules.getMetaData();
      int ColumnsC = RulesMeta.getColumnCount();
      String[] ColumnLabels = new String[ColumnsC];
      String[] ColumnValues = new String[ColumnsC];
      if (LDebug)
         System.err.print("generateRules: Loaded column [");
      for (int i=1; i <= ColumnsC;i++) {
         ColumnLabels[i-1]=RulesMeta.getColumnName(i);
         if (LDebug)
            System.err.print(", " + ColumnLabels[i-1]);
      } //for
      if (LDebug)
         System.err.println("] from table.");
      while(Rules.next()) {
         if (LDebug)
            System.err.println("generateRules: Loaded a row. Reading ["+ColumnsC+"] columns.");
         for (int i=1; i <= ColumnsC;i++) {
            if (Rules.getString(i) != null)
               ColumnValues[i-1] = Rules.getString(i);
            else
               ColumnValues[i-1] = new String();

         } //for
         // generate sql
         String SQLSOUPP = new String();
         String SQLSOUPN = new String();
         String SQLSOUPA = new String();
         boolean first = true;
         for (int i=1;i<=ColumnsC;i++) {
            if (ColumnValues[i-1].length() > 0) {
               if (first)
                  first = false;
               else {
                  SQLSOUPP = new String(SQLSOUPP + " AND ");
                  SQLSOUPN = new String(SQLSOUPN + " AND ");
                  SQLSOUPA = new String(SQLSOUPA + " AND ");
               } //if
               SQLSOUPP = new String(SQLSOUPP + ColumnLabels[i-1] + "='" + ColumnValues[i-1] +"' ");
               if (ColumnLabels[i-1].equals(ClassAttribute)) {
                  SQLSOUPN = new String(SQLSOUPN + ColumnLabels[i-1] + "<>'" + ColumnValues[i-1] +"' ");
                  SQLSOUPA = new String(SQLSOUPA + ColumnLabels[i-1] + "<>'*UNLIKELY!*' ");
               }
               else {
                  SQLSOUPN = new String(SQLSOUPN + ColumnLabels[i-1] + "='" + ColumnValues[i-1] +"' ");
                  SQLSOUPA = new String(SQLSOUPA + ColumnLabels[i-1] + "='" + ColumnValues[i-1] +"' ");
               } // else
            } //if
         } // for
         SQLSOUP = new String("SELECT COUNT("+ClassAttribute+") FROM "+TableName+" WHERE ");
         SQLSOUPP = new String(SQLSOUP + SQLSOUPP);
         SQLSOUPN = new String(SQLSOUP + SQLSOUPN);
         SQLSOUPA = new String(SQLSOUP + SQLSOUPA);

         if (LDebug)
            System.err.println("generateRules: About to query for positive rules with SQLSOUP ["+SQLSOUPP+"].");
         java.sql.Statement Measure = TheDatabase.createStatement();
         java.sql.ResultSet poscount = Measure.executeQuery(SQLSOUPP);
         poscount.next();
         int PositiveHits = poscount.getInt(1);
         if (LDebug)
            System.err.println("generateRules: About to query for negative rules with SQLSOUP ["+SQLSOUPN+"].");
         poscount = Measure.executeQuery(SQLSOUPN);
         poscount.next();
         int NegativeHits = poscount.getInt(1);
         if (LDebug)
            System.err.println("generateRules: About to query for coverage rules with SQLSOUP ["+SQLSOUPA+"].");
         poscount = Measure.executeQuery(SQLSOUPA);
         poscount.next();
         int CoverageHits = poscount.getInt(1);

         // we have the negative and positive hits, now generate the entries in the final table.
         String Colum = new String();
         String Values = new String();         
         first = true;
         for (int i=1;i<=ColumnsC;i++) {
            if (ColumnValues[i-1].length() > 0) {
               if (first)
                  first = false;
               else {
                  Colum = new String(Colum + ", ");
                  Values = new String(Values + ", ");
               } //if
               Colum = new String(Colum + ColumnLabels[i-1]);
               Values = new String(Values + "'"+ColumnValues[i-1]+"'");
            } //if
         } // for
         SQLSOUP = new String("INSERT INTO "+temptable+" ( "+Colum+", POSITIVE, NEGATIVE, COVERAGE ) VALUES ( "+Values+", "+PositiveHits+", "+NegativeHits
         +", " +CoverageHits+" )");
         if (LDebug)
            System.err.println("generateRules: Trying to insert a rules into ["+temptable+"] with SQL soup ["+SQLSOUP+"].");
         Measure.executeUpdate(SQLSOUP);
      } // while there are rules.     
      this.deleteTable(_table);
      SQLSOUP = new String("ALTER TABLE "+temptable+" RENAME TO "+_table);
      if (LDebug)
         System.err.println("generateResults: Renaming table ["+temptable+"] to table ["+_table+"] using SQL SOUP ["+SQLSOUP+"].");
      mystate.executeUpdate(SQLSOUP);
   } //generateResultsTable
   /***
     * Return a vector of String's containing all of the rules printed in a human-readable format.
     *
     * @param upper Set to true if the UpperBound rules are desired, false for the LowerBound.
     * @return A vector of Strings, one per rule.
     * @exception java.sql.SQLException Thrown if the jdbc has a problem.
    ***/
   public java.util.Vector getRules(boolean Upperbound) throws java.sql.SQLException {
      java.util.Vector thelist = new java.util.Vector();
      boolean LDebug = false;
      java.sql.Statement mystate = TheDatabase.createStatement();
      String thetable;
      if (Upperbound)
         thetable = RuleTableH;
      else
         thetable = RuleTableL;
      String SQLSOUP = new String("Select * from " + thetable);
      java.sql.ResultSet Rules = mystate.executeQuery(SQLSOUP);
      java.sql.ResultSetMetaData RulesMeta = Rules.getMetaData();
      int ColumnsC = RulesMeta.getColumnCount();
      String[] ColumnLabels = new String[ColumnsC];
      String[] ColumnValues = new String[ColumnsC];
      if (LDebug)
         System.err.print("getRules: Loaded column [");
      int classpos = 0;
      for (int i=1; i <= ColumnsC;i++) {
         ColumnLabels[i-1]=RulesMeta.getColumnName(i);
         if (ColumnLabels[i-1].equals(ClassAttribute))
            classpos = (i-1);
         if (LDebug)
            System.err.print(", " + ColumnLabels[i-1]);
      } //for
      if (LDebug)
         System.err.println("] from table.");
      while(Rules.next()) {
         if (LDebug)
            System.err.println("getRules: Loaded a row. Reading ["+ColumnsC+"] columns.");
         for (int i=1; i <= ColumnsC;i++) {
            if (Rules.getString(i) != null)
               ColumnValues[i-1] = Rules.getString(i);
            else
               ColumnValues[i-1] = new String();
         } //for
         String OneRule = new String();
         boolean first = true;
         for (int i=1; i <= ColumnsC;i++) {
            if (   (i != (classpos+1)) 
               &&  (!ColumnLabels[i-1].toUpperCase().equals("POSITIVE")) 
               &&  (!ColumnLabels[i-1].toUpperCase().equals("NEGATIVE"))
               &&  (!ColumnLabels[i-1].toUpperCase().equals("COVERAGE"))
               &&  (ColumnValues[i-1].length() >0) ) {
               if (first) 
                  first = false;
               else
                  OneRule= new String(OneRule + " AND ");
               OneRule = new String(OneRule + ColumnLabels[i-1] + "='"+ColumnValues[i-1] +"'");
            }//if
         } //fpr
         OneRule = new String("IF " + OneRule + " THEN " + ColumnLabels[classpos]+ "='" + ColumnValues[classpos] +"' ");
         OneRule = new String(OneRule + " which covers " + Rules.getString("POSITIVE") + " instances out of " 
         + Rules.getString("COVERAGE"));
         if (Rules.getInt("NEGATIVE") > 0)
            OneRule = new String(OneRule + " with " + Rules.getString("NEGATIVE") + " wrong classes.");
         else
            OneRule = new String(OneRule +".");
         if (LDebug)
            System.err.println("getRules: This rule looks like ["+OneRule+"].");
         thelist.addElement(OneRule);
      }//while
      return(thelist);
   } //getRules
}//class   

